#!/usr/bin/perl -w
#< Boilerplate
#
# The Aranea question answering system
# Copyright (C) 2002-2005, Jimmy Lin
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, please visit
# http://www.gnu.org/copyleft/gpl.html
#
# $Header: /fs/clip-qa/.cvsroot//Aranea/pm/Aranea/Modules/BoostClosedClassAnswers.pm,v 1.14 2005/06/10 16:17:16 jimmylin Exp $
#
#>
#< POD
###############################################################################

=head1 NAME

B<C<Aranea::Modules::BoostClosedClassAnswers>> - Aranea module for
boosting closed class answers.

=head1 SYNOPSIS

This Aranea modules is meant to be used in conjunction with the
L<C<Aranea::Shell>|Aranea::Shell>.

=head1 DESCRIPTION

This Aranea module ensures that closed class answers are valid,
throwing away any invalid candidates.  A closed classed answer is one
whose form can be either exhaustively enumerated or easily detected.
For example, the answer to a "what state..." question must be a state,
and since there are only 50 states, the module can check to see if the
answer candidate is a valid state.  "What president..." and "what
country..." questions are two more classes than fall under this
category.  Another type of closed class answer involves extracting the
question focus, e.g., the answer to a "how many feet..." question must
have the unit `feet', so the module can verify correct units and throw
out all invalid forms.

=cut

###############################################################################
#>
#< Header
package Aranea::Modules::BoostClosedClassAnswers;

use strict;
use warnings;

use Data::Dumper;
use IO::Handle;

use Aranea::XML::Aranea;
use Aranea::Patterns::Dates qw($MONTH);
use Aranea::Patterns::Locations qw($COUNTRY);
use Aranea::Patterns::Persons qw($US_PRESIDENT);

our %WORLD_CITIES;

#>

#< STDIN Handler
unless ( caller() ) {
    my $stdin = new_from_fd IO::Handle(fileno(STDIN),"r");
    my $Aranea = Aranea::XML::Aranea->new($stdin);

    $Aranea = doModule($Aranea);

    $Aranea->write_xml();
}
#>

#< sub doModule
sub doModule {
    my $Aranea = shift;

    # July 31, 2003 -- divided 'BoostClosedClassAnswers' into
    # 'FilterClosedClass' and 'BoostClosedClassAnswers'
    #
    # "Closed-class" answers can actually be divided into two types:
    # ones which we can enumerate (Type 1) and ones that we can easily
    # recognize, but not enumerate (Type 2).  Countries and states are
    # examples of the Type 1 answer, and age is a good example of the
    # Type 2 answer.
    #
    # The insight is to put all Type 1 fixed lists into a filter that
    # runs before all the n-gram collection code.  If we have a
    # exhaustive list, then we don't need redundancy to assemble
    # larger answers from smaller answer fragments.  In such cases,
    # the n-gram collection process will often make the answer worse
    # (by tacking junk on either side, etc.)
    #
    # Answers whose form we can readily recognize, but cannot
    # enumerate (Type 2) are best "boosted" after the n-gram
    # collection process.  This allows us to build up longer answers
    # and then filter them out.  A great example is something like
    # "How old is the sun?".  A typical correct answer like "4.5
    # billion years" is relatively rare in its exact form... due to
    # the Google snippets, we are more likely to get fragments like
    # "4.5", "4.5 billion", etc.  Letting the data redundancy/n-gram
    # collection code do its job gives the system a chance to assemble
    # the longer answer (recall).  Applying a filter to recognized the
    # answer type then brings back our precision.

    $Aranea->insert_pipeline_tag('$Id: BoostClosedClassAnswers.pm,v 1.14 2005/06/10 16:17:16 jimmylin Exp $ ');

    # even if there are no requests, there may be work to do,
    # e.g., q387 "What year did Montana become a state?":
    # the answer is retrieved by database lookup via 50states.com,
    # but we want to extract only the year.

    my $query = $Aranea->query()->text();
    my $entries = $Aranea->candidate_list()->entries();

    # Don't do number/unit filtering if answer has already be
    # retrieved via database lookup, because the database answer might
    # contain a little bit more information, e.g., '1 mile (1.6 km)',
    # which would not pass the filter.  The only downside of this is
    # that if the question asks for specific units, e.g., 'How far in
    # miles is...', then we're screwed: if we give both answers, it's
    # not an 'exact' answer; on the other hand, being able to do unit
    # conversions on the fly requires lots of smarts that we don't
    # have (yet).
    if ( $Aranea->request_list()->entries_count() != -1 ) {
	# filter against ways of writing age
	$Aranea->candidate_list()->filter_candidates(\&match_age)
	    if $query =~ /^How old/i or
		$query =~ /^What is the age/i;

	# filter against ways of writing length
	$Aranea->candidate_list()->filter_candidates(\&match_length_unit)
	    if $query =~ /^How tall/i or
		$query =~ /^How far/i or
		$query =~ /^How wide/i or
	       ( $query =~ /^What is/ and
		 $query =~ /(diameter|depth|distance|length|width|height)/ ) or
               ( $query =~ /^How big /i and
		 $query =~ /(diameter|depth|distance|length|width|height)/ );

	# With "How many X" type questions, the answer should either
	# be a pure number, or a number with unit X (which is the
	# question focus).  However, don't bother if the answer was
	# retrieved via database lookup.

	my $qfocus = $Aranea->query()->qfocus();
	if ( $query =~ /^How many /i and $qfocus ) {
	    my $criteria = 
		sub { shift =~ /(^[0-9,.]+$| $qfocus$)/ };
	    $Aranea->candidate_list()->filter_candidates($criteria);
	}
    }

    #print Dumper($Aranea->candidate_list()->entries());

    return $Aranea;
}
#>

#< match_age
sub match_age {
    my $s = shift;

    return '' if $s =~ /^1\d\d\d$/;

    $s =~ /^\d+ years/ or
	$s =~ /^\d+$/ or     # a pure number
        $s =~ /^[.\d]+ (million|billion|trillion)( years)?$/;
        # some very large number, e.g., age of the sun
}
#>
#< match_length_unit
sub match_length_unit {
    my $s = shift;

    $s =~ / (inches|in|
             feet|ft|
             yards|
             miles|mi|
             nanometers|nm|
             microns|
             micrometers|
             millimeters|mm|
             centimeters|cm|
             meters|m|
             kilometers|km|
             AU|
             light\s+years)$/xi
    and $s =~ /^\d+/;
}
#>

1;
