#!/usr/bin/perl -w
#< Boilerplate
#
# The Aranea question answering system
# Copyright (C) 2002-2005, Jimmy Lin
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, please visit
# http://www.gnu.org/copyleft/gpl.html
#
# $Header: /fs/clip-qa/.cvsroot//Aranea/pm/Aranea/Modules/GenerateNGrams.pm,v 1.29 2005/06/10 16:17:16 jimmylin Exp $
#
#>
#< POD
###############################################################################

=head1 NAME

B<C<Aranea::Modules::GenerateNGrams>> - Aranea module for generating
ngrams from raw answer candidates

=head1 SYNOPSIS

This Aranea modules is meant to be used in conjunction with the
L<C<Aranea::Shell>|Aranea::Shell>.

=head1 DESCRIPTION

This Aranea module generates unigrams, bigrams, trigrams, and
quadgrams from each candidate in the C<candidate_list>.  The original
C<candidate_list> is then discarded and replaced by the new n-grams.
This module also performs some basic filtering of the candidates,
discarding any that contain certain special symbols; see actual source
code and comments therein for details.

=cut

###############################################################################
#>
#< Header
package Aranea::Modules::GenerateNGrams;

use strict;
use warnings;
use lib qw(lib pm);                         # needed for command-line mode

use IO::Handle;

use Aranea::XML::Aranea;
use Aranea::XML::CandidateList;
use Aranea::XML::CandidateEntry;
use Aranea::Patterns::Locations qw($POSTAL_CODE);

our %STOPWORDS = (
  'AD' => 1,
  'BC' => 1,
  'Page' => 1,
  '2001' => 1,
  '2002' => 1,
  '2003' => 1,
  'PO Box' => 1,
);

#>

#< STDIN handler
unless ( caller() ) {
    my $stdin = new_from_fd IO::Handle(fileno(STDIN),"r");
    my $Aranea = Aranea::XML::Aranea->new($stdin);

    $Aranea = doModule($Aranea);

    $Aranea->write_xml();
}
#>

#< sub doModule
sub doModule {
    my $Aranea = shift;

    $Aranea->insert_pipeline_tag('$Id: GenerateNGrams.pm,v 1.29 2005/06/10 16:17:16 jimmylin Exp $ ');

    # if there are no requests, then there is nothing to do
    return $Aranea if $Aranea->request_list()->entries_count() == -1;

    my $entries = $Aranea->candidate_list()->entries();
    my $new_entries = [];

    for my $entry (@$entries) {
	my $candidate = $entry->candidate();

	# ignore canidates that have vertical bars --- often, they are
	# used as navigation aids, e.g., "Help | Home | About"
	next if $candidate =~ /(^| )\|($| )/;

	# Kill all leading and trailing special characters
	$candidate =~ s/^\W+//g;
	$candidate =~ s/\W+$//g;

	my @ngrams = _generate_all_ngrams(4, $candidate);

	foreach my $gram (@ngrams) {
	    # kill all candidate with special symbols
	    next if $gram =~ m/[\/\=():;?]/;
	    next if $gram =~ m/''/;
	    next if $gram =~ m/``/;
	    next if $gram =~ m/ -+ /;

	    # ignore state, zip combinations (should be in a better
	    # place?)
	    next if $gram =~ m/^$POSTAL_CODE \d\d\d\d\d$/o;

	    # ignores astronomical units (definitely should be in a
	    # better place...)
	    next if $gram =~ m/^[\d.]+ AU$/;

	    push @$new_entries,
         	 Aranea::XML::CandidateEntry->new($gram, $entry->score());
	}
    }

    $Aranea->candidate_list()->entries($new_entries);

    return $Aranea;
}
#>

#< sub _generate_all_ngrams
sub _generate_all_ngrams {
    my ($n, $candidate) = @_;

    # split along whitespace
    my @grams = split(/\s+/, $candidate);

    my $gram_rank = ($n >= @grams) ? @grams : $n;
    my @result = ();

    for (my $i = 0; $i < $gram_rank; $i++) {
	# generate all i-grams
	# start on the j'th word
	for (my $j = 0; $j < @grams - $i; $j++) {
	    my $gram = "$grams[$j]";
	    # use i elements
	    for (my $k = 1; $k <= $i; $k++) {
		my $w = $j+$k;
		$gram = "$gram $grams[$j+$k]";
	    }

	    $gram =~ s/^\W+//g;
	    $gram =~ s/\W+$//g;

	    # kill possessive endings
	    $gram =~ s/'s$//g;

	    $gram =~ s/&#39;/'/g;

	    # four digit numbers are often written without the comma:
	    # normalize the candidate in this case.  Note that we want
	    # to remove the comma instead of adding it; adding commas
	    # would screw up all four digit years.
	    $gram =~ s/(^| )(\d),(\d\d\d)( |$)/$1$2$3$4/;

	    # Make sure that candidates are not in the "special
	    # stopwords" list.  Note that we do this after stripping
	    # the leading and trailing punctuation/special characters.
	    push(@result, $gram) unless $STOPWORDS{$gram};
	}
    }

    return @result;
}
#>

1;
