#!/usr/bin/perl -w
#< Boilerplate
#
# The Aranea question answering system
# Copyright (C) 2002-2005, Jimmy Lin
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, please visit
# http://www.gnu.org/copyleft/gpl.html
#
# $Header: /fs/clip-qa/.cvsroot//Aranea/pm/Aranea/Modules/GenerateReformulations.pm,v 1.17 2005/06/10 16:17:16 jimmylin Exp $
#
#>
#< POD
###############################################################################

=head1 NAME

B<C<Aranea::Modules::GenerateReformulations>> - Aranea module for
generating reformulation requests

=head1 SYNOPSIS

This Aranea modules is meant to be used in conjunction with the
L<C<Aranea::Shell>|Aranea::Shell>.

=head1 DESCRIPTION

This Aranea module populates the C<request_list> with various
reformulations based on analysis of the question.

=cut

###############################################################################
#>
#< Header
package Aranea::Modules::GenerateReformulations;

use strict;
use warnings;
use lib qw(lib pm);                         # needed for command-line mode

use IO::Handle;

use Aranea::XML::Aranea;
use Aranea::Lexicon::Lexicon qw(:default);

our $inexact_snippets_count = 100;
our $exact_snippets_count = 100;

#>

#< STDIN handler
unless ( caller() ) {
    my $stdin = new_from_fd IO::Handle(fileno(STDIN),"r");
    my $Aranea = Aranea::XML::Aranea->new($stdin);

    $Aranea = doModule($Aranea);

    $Aranea->write_xml();
}
#>

#< sub doModule
sub doModule {
    my $Aranea = shift;

    $Aranea->insert_pipeline_tag('$Id: GenerateReformulations.pm,v 1.17 2005/06/10 16:17:16 jimmylin Exp $ ');

    # if we have the answer already (e.g., via database lookup), then
    # don't bother generating requests.
    return $Aranea if $Aranea->candidate_list()->entries_count() != -1;

    my $qobj = $Aranea->query();
    my $q = $qobj->text();
    my $tagq = $qobj->tagged();
    my $original = $qobj->original();
    my $request_list = $Aranea->request_list();

    # just in case there is a separately tagged punctation...  there
    # shouldn't be...
    $tagq =~ s/\?\/\.\s+$//g;

    # NOTE: the quoted pattern is not mutally exlusive with the other
    # patterns
    if ( $q =~ /\"/i ) {
	map { $request_list
		  ->add_default_inexact_query($_, $inexact_snippets_count) }
	    expand_quoted_inexact($q, $tagq);
    }

    if ( $q =~ /^How old /i ) {
	map { $request_list
		  ->add_default_inexact_query($_, $inexact_snippets_count) }
	    expand_how_old_inexact($q, $tagq);
    }
    elsif ( $q =~ /^Where /i ) {
	map { $qobj->add_reformulation($_) }
	    expand_where_exact($q, $tagq);

	map { $request_list
		  ->add_default_inexact_query($_, $inexact_snippets_count) }
	    expand_where_inexact($q, $tagq);
    }
    elsif ( $q =~ /^When /i ) {
	my @exact = expand_when_exact($tagq);
	map { $qobj->add_reformulation($_) } @exact;
    }
    elsif ( $q =~ /^Who /i ) {
	my @exact = expand_who_exact($q, $tagq);
	map { $qobj->add_reformulation($_) } @exact;

	map { $request_list
		  ->add_default_inexact_query($_, $inexact_snippets_count) }
	    expand_who_inexact($q, $tagq);
    }
    elsif ( $q =~ /^What /i ) {
	my @exact = expand_what_exact($q, $tagq);
	map { $qobj->add_reformulation($_) } @exact;

	map { $request_list
		  ->add_default_inexact_query($_, $inexact_snippets_count) }
	    expand_what_inexact($q, $tagq);

	# One would think that droping the question focus, e.g.,
	# 'city' in 'What city' would improve performance... however,
	# it doesn't.  On 125 questions 'What NP' questions from
	# TREC9/10, a run got a MRR of 0.500 (56 at rank 1) without
	# this following code segment and a MRR of 0.484 (54 and rank
	# 1) with the code segment.
    }
    else {
	my @exact = expand_other_exact($original, $q, $tagq);
	map { $qobj->add_reformulation($_) } @exact;

	map { $request_list
		  ->add_default_inexact_query($_, $inexact_snippets_count) }
	    expand_other_inexact($original, $q, $tagq);
    }

    my $tensedq = expand_tense($tagq);
    if ( $tensedq ) {
	$qobj->add_reformulation("$tensedq ?x");
	$Aranea->request_list()
	    ->add_default_inexact_query($tensedq, $inexact_snippets_count);
    }

    map { $request_list
	      ->add_default_exact_query($_, $exact_snippets_count) }
	@{$Aranea->query()->get_reformulations()}
	    if $Aranea->query()->get_reformulations();

    return $Aranea;
}
#>

#< expand_quoted_inexact
sub expand_quoted_inexact {
    my ($q, $tagq) = @_;

    my @requests = ();

    my @quoted = ();

    while ( $q =~ /"(.*?)"/g ) {
	push @quoted, $1;
    }

    foreach my $t ( @quoted ) {
	$q =~ s/"\Q$t\E"//g;
	$t =~ s/ / \+/g;
	$t =~ s/^/+/;
	$t = "\"$t\"";
    }
    $q =~ s/\W+$//g;
    $q =~ s/ +/ /g;

    my $reform = join(' ', @quoted) . " $q";
    #print "==$reform\n";
    #$q =~ s/^How old //i;
    #$q =~ s/\W+$//g;

    #my $reform = "$q \"years old\"";
    #print "==$reform==\n";

    push @requests, $reform;

    return @requests;
}
#>

#< expand_how_old_inexact
sub expand_how_old_inexact {
    my ($q, $tagq) = @_;

    my @requests = ();

    $q =~ s/^How old //i;
    $q =~ s/\W+$//g;

    my $reform = "$q \"years old\"";
    #print "==$reform==\n";

    push @requests, $reform;

    return @requests;
}
#>

#< expand_where_exact
sub expand_where_exact {
    my ($q, $tagq) = @_;

    my @requests = ();

    if ( $q =~ /^Where (is|are|was) (.*)/i ) {
	my ($bev, $n) = ($1, $2);

	if ( $n =~ /^there / and $tagq =~ /there\/EX/ ) {
	    $n =~ s/[^a-z]+$//g;
	    $n =~ s/^there //g;

	    my $reformulation = "there $bev $n in ?x";
	    push @requests, $reformulation;
	} elsif ( $tagq =~ /\/VB.? $/ ) {
	    # Exact rewrites for questions like q1326 "Where are the
	    # British crown jewels kept?"

	    $n =~ s/ ([^ ]+)$//g;
	    my $v = $1;
	    $v =~ s/[^a-z]+$//g;

	    my $reformulation = "$n $bev $v in ?x";
	    push @requests, $reformulation;
	} else {
	    # Exact rewrite for "Where is X?" questions; watch out for
	    # both singular, e.g., q1211 "Where is the Little League
	    # Museum?", but also plural, e.g., q1036 "Where are the
	    # National Archives?"

	    $n =~ s/[^a-z]+$//g;

	    my $reformulation = "$n $bev located in ?x";
	    push @requests, $reformulation;
	}
    }

    return @requests;
}
#>
#< expand_where_inexact
sub expand_where_inexact {
    my ($q, $tagq) = @_;

    my @requests = ();

    if ( $q =~ /^Where (is|are|was) (.*)/i ) {
	my ($bev, $n) = ($1, $2);

	if ( $n =~ /^there / and $tagq =~ /there\/EX/ ) {
	    $n =~ s/[^a-z]+$//g;
	    $n =~ s/^there //g;
	    $n =~ s/ / \+/g;
	    $n =~ s/^/+/;

	    my $reformulation = "\"$n\"";
	    push @requests, $reformulation;
	} elsif ( $tagq =~ /\/VB.? $/ ) {
	    # Inexact rewrites for questions like q1326 "Where are the
	    # British crown jewels kept?"

	    $n =~ s/ ([^ ]+)$//g;
	    my $v = $1;
	    $v =~ s/[^a-z]+$//g;

	    $n =~ s/ / \+/g;
	    $n =~ s/^/+/;

	    # The following reformulation decreases performance:
	    #     "$n" "+$bev $v +in"
	    #   -- July 28, 2003, jimmylin

	    push @requests, "\"$n\"";
	} else {
	    # Exact rewrite for "Where is X?" questions; watch out for
	    # both singular, e.g., q1211 "Where is the Little League
	    # Museum?", but also plural, e.g., q1036 "Where are the
	    # National Archives?"

	    $n =~ s/[^a-z]+$//g;

	    $n =~ s/ located$//g;
	    $n =~ s/ / \+/g;
	    $n =~ s/^/+/;

	    # The following reformulation decreases performance:
	    #     "$n" "+$bev located +in"
	    #   -- July 28, 2003, jimmylin

	    push @requests, "\"$n\"";
	}
    }

    return @requests;
}
#>

#< expand_when_exact
sub expand_when_exact {
    my $tagq = shift;

    my @requests = ();

    if ( $tagq =~ /^When\/WRB was\/VBD (.*) ([^\/]+\/(VBN|VBD))/ ) {
	my ($np, $v) = ($1, $2);
	$np =~ s/\/[^ ]+//g;
	$v =~ s/\/[^ ]+//g;
	push @requests, "$np was $v ?x";

	push @requests, "$np was founded ?x" if $v eq 'established';
    }

    return @requests;
}
#>

#< expand_other_exact
sub expand_other_exact {
    my ($original, $q, $tagq) = @_;
    my @requests = ();

    return @requests;
}
#>
#< expand_other_inexact
sub expand_other_inexact {
    my ($original, $q, $tagq) = @_;
    my @requests = ();

    my %hash = ( 'tall' => 'height',
		 'wide' => 'width',
		 'deep' => 'depth',
		 'old'  => 'age'
	       );

    if ( $q =~ /^How (tall|wide|deep) /i ) {
	my $nform = $hash{$1};
	my $reform = "$nform " . $';
	$reform =~ s/\W+$//g;

	push @requests, $q . " $nform";
    }

    if ( $q =~ /^How (tall|wide|deep) is /i ) {
	my $nform = $hash{$1};
	my $reform = "$nform of " . $';
	$reform =~ s/ / \+/g;
	$reform =~ s/\W+$//g;
	$reform = '"+' . $reform . '"';

	push @requests, $reform;
    }

    if ( $q =~ /^How far /i ) {
	my $reform = "distance " . $';

	push @requests, $reform;
    }

    # the following reformation doesn't work: 
    #   'Name a flying mammal' -> "+flying +mammal"

    return @requests;
}
#>

#< expand_who_exact
sub expand_who_exact {
    my ($q, $tagq) = @_;

    my @requests = ();

    #print "$tagq\n";
    #print "$n\n";

    if ( $tagq =~ /^Who\/WP ([^\/]+)\/VB[NDZ] (.*)/ ) {
	my ($rest, $v) = ($2, $1);

	#unless ( $v =~ /(is|was)/ ) {

	#print "==$rest\n";
	    $rest =~ s/^[^\/]+\/DT [^\/]+\/NN (?=")//;

	    $rest =~ s/\W$//g;
	    $rest =~ s/\/[^ ]+//g;

	    my $stems = stem($v);
	    #map { print " [$_->[0], $_->[1], $_->[2]]\n" } @$stems;
	    my $stemmed_v = '';
	    foreach ( @$stems ) {
		$stemmed_v = $_->[0], last if $_->[1] eq 'v';
	    }
	    $stemmed_v = $v unless $stemmed_v;
	    my $infl = inflect_verb($stemmed_v);

	    my $inflected_v = '';
	    foreach ( @$infl ) {
		$inflected_v = $_->[1], last if $_->[0] eq 'pa';
	    }
	    $inflected_v = $v unless $inflected_v;
	    #print "##$inflected_v\n";

	    push @requests, "?x $v $rest";
	    push @requests, "$rest $v ?x" if $v =~ /is|was/;
	    push @requests, "$rest was $inflected_v by ?x" if $v !~ /is|was/;
	#}
    }

    #print "==$tagq\n";
    #if ( $tagq =~ /^Who\/WP / ) { #([^\/]+)\/VB[NDZ]/ ) {
	#print "##$tagq\n";
    #}

    return @requests;
}
#>
#< expand_who_inexact
sub expand_who_inexact {
    my ($q, $tagq) = @_;

    my @requests = ();

    if ( $q =~ /^Who (is|was) (.*)/i ) {
	my $n = $2;

	$n =~ s/[^a-z]+$//g;
	$n =~ s/ / \+/g;      # add + in front of every token (for Google)
	$n =~ s/^/+/;

	my $reform = "\"$n\"";
	push @requests, $reform;
    }

    return @requests;
}
#>

#< expand_what_exact
sub expand_what_exact {
    my ($q, $tagq) = @_;
    my @requests = ();

    # exact reformulations for 'made of' questions
    # i.e., q1093, "What is pastrami made of?"
    #       q1119, q1146, q1305, q1173
    #
    # generate the query "pastrami is made of ?x"

    if ( $tagq =~ /^What\/[^\s]+.*?                 # What, potential NP
                    (is|are)\/[^\s]+\s
                    (.*?)\s                         # the main NP
                    ([^\/]+)\/VB.?\s([^\/]+)\/IN    # main Verb and prep
                  /x ) {

        my $reform = "$2 $1 $3 $4 ?x";
	$reform =~ s/\/[^ ]+//g;

	#print "$reform\n";
	push @requests, $reform;
    }

    return @requests;
}
#>
#< expand_what_inexact
sub expand_what_inexact {
    my ($q, $tagq) = @_;
    my @requests = ();

    # inexact reformulations for 'made of' questions
    # i.e., q1093, "What is pastrami made of?"
    #       q1119, q1146, q1305, q1173
    #
    # generate the query 'pastrami "+made +of"'

    if ( $tagq =~ /^What\/[^\s]+.*?                 # What, potential NP
                    (is|are)\/[^\s]+\s
                    (.*?)\s                         # the main NP
                    ([^\/]+)\/VB.?\s([^\/]+)\/IN    # main Verb and prep
                  /x ) {

	my $reform = "$2 \"+$3 +$4\"";
	$reform =~ s/\/[^ ]+//g;

	#print "$reform\n";
	push @requests, $reform;
    }

    if ( $q =~ /What (are|is|was) /i and $tagq =~ /\/VB.?/ ) {
	(my $np = $q ) =~ s/What (are|is|was) //g;

	$np =~ s/ / +/g;
	$np =~ s/^/+/;
	$np =~ s/\?$//;

	push @requests, $np;
    }

    return @requests;
}
#>

#< expand_tense
sub expand_tense {
    my $tagq = shift;

    if ( $tagq =~ / did\/VBD / ) {
	#print "$tagq\n";

	$tagq =~ s/\?\/\.\s+$//g;

	my @tkns = split(' ', $');    # the input tokens
	my @qtkns = ();               # for the output

	# sometimes the tagger screws up and doesn't even find a main
	# verb in the question, e.g.,  Q375,
	#
	#    What/WP ocean/NN did/VBD the/DT Titanic/NNP sink/NN in/IN ?/.
	#
	# this flag notes if _any_ token has been tagged as a verb
	my $no_verb = 0;
	$no_verb = 1 if $' !~ /\/VB/;

	# this flag notes if a verb has been encountered in the input
	# token sequence
	my $has_verb = 0;

	# iterate through all the tokens
	for my $i ( 0 .. $#tkns ) {
	    my $t = $tkns[$i];
	    next unless $t =~ /\/[A-Z]/;

	    # gets the actual word and its POS
	    my ($w, $pos) = split ('/', $t);

	    #print "before: $w\n";
	    $w =~ s/^\W+//;
	    $w =~ s/\W+$//;
	    #print "after: $w\n";

	    # to handle questions like
	    #   q422, When did Princess Diana and Prince Charles get married?
	    #   q928, When did John F. Kennedy get elected as President?
	    # it appears that throwing out 'get' leads to better performance
	    next if $w eq 'get';

	    # we want to inflect if:
	    #  1. it's a verb
	    #  2. it's the last token and no verb has been encountered
	    #  3. there isn't even a word tagged as a verb (and the POS
            #     isn't a NNP, e.g., 'World War I' shouldn't become
            #     'World Warred I'
	    if ( $pos =~ /^VB/ or
		 ( not $has_verb and $i == $#tkns ) or
		 ( $no_verb and $pos ne 'NNP' )) {
		my $w_infl = get_past_tense($w);
		push @qtkns, $w_infl;
		$has_verb = 1;
	    } else {
		# otherwise, just keep the token
		push @qtkns, $w;
	    }
	}

    	return join(' ', @qtkns);
    }

    return;
}
#>
#< get_past_tense
sub get_past_tense {
    my $w = shift;

    my $infl = inflect_verb($w);
    return $w unless $infl;

    #map { print " [$_->[0], $_->[1], $_->[2]]\n" } @$infl;

    foreach my $t ( @$infl ) {
	return $t->[1] if $t->[0] eq 'aP';
    }

    return $w;
}
#>

#< set_inexact_snippets_count
sub set_inexact_snippets_count {
    $inexact_snippets_count = shift;
}
#>
#< set_exact_snippets_count
sub set_exact_snippets_count {
    $exact_snippets_count = shift;
}
#>

1;
