#!/usr/bin/perl -w
#< Boilerplate
#
# The Aranea question answering system
# Copyright (C) 2002-2005, Jimmy Lin
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, please visit
# http://www.gnu.org/copyleft/gpl.html
#
# $Header: /fs/clip-qa/.cvsroot//Aranea/pm/Aranea/Modules/NormalizeCandidates.pm,v 1.13 2005/06/10 16:17:16 jimmylin Exp $
#
#>
#< POD
###############################################################################

=head1 NAME

B<C<Aranea::Modules::NormalizeCandidates>> - Aranea module for
normalizing answer candidates.

=head1 SYNOPSIS

This Aranea modules is meant to be used in conjunction with the
L<C<Aranea::Shell>|Aranea::Shell>.

=head1 DESCRIPTION

This Aranea module normalizes specific types of answer candidates
(e.g., dates).

=cut

###############################################################################
#>
#< Header
package Aranea::Modules::NormalizeCandidates;

use strict;
use warnings;

use Data::Dumper;
use IO::Handle;

use Aranea::Patterns::Dates qw($MONTH $MONTH_ABBREV %ABBREV_TO_MONTH);
use Aranea::Patterns::Locations qw(exactify_location $ADJ_TO_LOCATION_keys $POSTAL_CODE %POSTAL_CODE_TO_STATE);
use Aranea::Patterns::Persons qw($TITLE $OCCUPATION);
use Aranea::XML::Aranea;

#>

#< STDIN Handler
unless ( caller() ) {
    my $stdin = new_from_fd IO::Handle(fileno(STDIN),"r");
    my $Aranea = Aranea::XML::Aranea->new($stdin);

    $Aranea = doModule($Aranea);

    $Aranea->write_xml();
}
#>

#< sub doModule
sub doModule {
    my $Aranea = shift;

    $Aranea->insert_pipeline_tag('$Id: NormalizeCandidates.pm,v 1.13 2005/06/10 16:17:16 jimmylin Exp $ ');

    # even if there are no requests, we want normalize knowledge
    # annotation results

    my $query = $Aranea->query()->text();
    my $tagq = $Aranea->query()->tagged();

    $Aranea->candidate_list()->map_candidates(\&normalize_general);

    $Aranea->candidate_list()
	->map_candidates(\&normalize_date)
	    if $query =~ /^When /i;

    $Aranea->candidate_list()
	->map_candidates(\&normalize_location)
	    if $query =~ /^Where /i;

    $Aranea->candidate_list()
	->map_candidates(\&normalize_who)
	    if $query =~ /^Who /i;

    #print Dumper($Aranea->candidate_list()->entries());

    return $Aranea;
}
#>

#< sub normalize_general
sub normalize_general {
    my $s = shift;

    # Sometimes, we get the same answer repeated twice in the same
    # candidate.
    my $len = length($s);
    if ( $len % 2 == 1 ) {
	my $part1 = substr($s, 0, ($len-1)/2);
	my $part2 = substr($s, ($len+1)/2, ($len-1)/2);

	$s = $part1 if $part1 =~ /\Q$part2\E/;
    }

    # kill possessives
    $s =~ s/'s$//i;

    $s =~ s/(\d)ft/$1 ft/i;

    return $s;
}
#>
#< sub normalize_date
sub normalize_date {
    my $s = shift;

    # normalize '14 July' to 'July 14'
    $s =~ s/^(\d+)(?:st|nd|rd|th)? ($MONTH)/$2 $1/;

    # normalize 'July 14th-17th' to 'July 14-17'
    $s =~ s{^($MONTH)\s+(\d+)(?:st|nd|rd|th)?\s*-+\s*(\d+)(?:st|nd|rd|th)?}
           {$1 $2-$3}x;

    # normalize 'July 14th' to 'July 14'
    $s =~ s/^($MONTH) (\d+)(?:st|nd|rd|th)?/$1 $2/;

    # normalize 'March 14 1879' to 'March 14, 1879'
    $s =~ s/^($MONTH) (\d\d?) (\d+)/$1 $2, $3/;

    # expand abbreviation of months
    $s =~ s/( |^)($MONTH_ABBREV)(,| |$)/$1$ABBREV_TO_MONTH{$2}->[0]$3/;

    return $s;
}
#>
#< sub normalize_location

sub normalize_location {
    my $s = shift;

    $s =~ s/Washington DC/Washington, DC/g;
    $s =~ s/, ($POSTAL_CODE)$/, $POSTAL_CODE_TO_STATE{$1}->[0]/g;

    # normalize 'Saint Louis' to 'St. Louis'
    $s =~ s/^Saint /St. /;

    $s = exactify_location($s);

    return $s;
}

#>
#< sub normalize_who
sub normalize_who {
    my $s = shift;

    # kill leading nationality
    $s =~ s/^($ADJ_TO_LOCATION_keys) //i;

    # After we've killed off nationality, there may be other downcased
    # words to kill, e.g. "English clergyman John Newton"
    $s =~ s/^([a-z,\-]+ )+([A-Z])/$2/;

    # kill leading titles and occupations
    $s =~ s/^($TITLE|$OCCUPATION),?\s+//i;

    # sometimes locatives are included in the candidate answer
    $s =~ s/ (in|at) [A-Z]\w+$//;

    # take care of answers like "Ts'ai Lun, a Chinese"
    $s =~ s/, (an? |the )?($ADJ_TO_LOCATION_keys|$OCCUPATION|$TITLE)$//i;

    # exactify things like "Fiction author William Gibson"
    $s =~ s/^[A-Z][a-z]+ $OCCUPATION (?=[A-Z])//i;

    return $s;
}
#>

1;
