#!/usr/bin/perl -w
#< Boilerplate
#
# The Aranea question answering system
# Copyright (C) 2002-2005, Jimmy Lin
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, please visit
# http://www.gnu.org/copyleft/gpl.html
#
# $Header: /fs/clip-qa/.cvsroot//Aranea/pm/Aranea/Modules/ProjectAnswerOnTREC.pm,v 1.25 2005/06/10 16:17:16 jimmylin Exp $
#
#>
#< POD

###############################################################################

=head1 Name

B<C<Aranea::Modules::ProjectAnswerOnTREC>> - projects Aranea answers
onto the TREC documents.

=head1 Synopsis

This Aranea modules is meant to be used as part of a larger pipeline;
for an example, please consult C<src/answerit.pl>

=head1 Description

Answer projection for TREC-2002. (Tweaked for TREC 2003 and TREC
2004.)

=cut

###############################################################################

#>
#< Header
package Aranea::Modules::ProjectAnswerOnTREC;

use strict;
use warnings;

use IO::Handle;

use Aranea::XML::Aranea;
use Aranea::Common qw(:default);
use Aranea::Lexicon::Lexicon qw(:default);
use Aranea::Patterns::Dates qw($MONTH $MONTH_FULL $DAY $YEAR %MONTH_TO_ABBREV);
use Aranea::Patterns::Locations qw($STATE %STATE_TO_POSTAL_CODE %STATE_TO_ABBREV);
use Aranea::Patterns::Persons qw($FEMALE_FIRST_NAME $MALE_FIRST_NAME);

#>

#< STDIN handler
unless ( caller() ) {
    my $stdin = new_from_fd IO::Handle(fileno(STDIN),"r");
    my $Aranea = Aranea::XML::Aranea->new($stdin);

    $Aranea = doModule($Aranea);

    $Aranea->write_xml();
}
#>

#< sub doModule

sub doModule {
    my $Aranea = shift;

    $Aranea->insert_pipeline_tag('$Id: ProjectAnswerOnTREC.pm,v 1.25 2005/06/10 16:17:16 jimmylin Exp $ ');

    my $q = $Aranea->query()->text();
    my $qid = $Aranea->query()->id();

    #my @TRECdocs = get_documents_prise($qid);
    #my @TRECdocs = get_documents_lucene($qid);
    my @TRECdocs = get_documents_trec2004($qid);

    print "Obtained: $#TRECdocs documents.\n";

    return $Aranea if $Aranea->candidate_list()->entries_count() == -1;

    my $entries = $Aranea->candidate_list()->entries();

    my $new_entries = [];
    for my $entry ( @$entries ) {
	my $candidate = $entry->candidate();
	my ($support, $score, $new_ans) =
	    project($q, $candidate, \@TRECdocs);

	$entry->candidate($new_ans)
	    if $new_ans and $new_ans ne $candidate;
	$entry->projection_score($score);
	$entry->{support}->{doc} = [ $support ];

	if ( $support ) {
	    push @$new_entries, $entry;
	}
    }

    $Aranea->candidate_list()->entries($new_entries);
    $entries = $new_entries;

    my $max_score = 1;
    for my $entry ( @$entries ) {
	my $score = $entry->projection_score();

	$max_score = $score if $score > $max_score;
    }

    #print "max projection score = $max_score\n";
    for my $entry ( @$entries ) {
	my $score = $entry->projection_score();

	$entry->projection_score($score/$max_score);
    }

    for my $entry ( @$entries ) {
	my $pscore = $entry->projection_score();
	my $score = $entry->score();
	my $candidate = $entry->candidate();

	$entry->score($score + $pscore);
	#print "##$candidate##$score##$pscore##\n";
    }

    return $Aranea;
}

#>

#< get_documents_prise
sub get_documents_prise {
    my $qno = shift;

    my $topdocs = '';

    open(TOPDOC, "< /projects/infolab-ir/trec2003_prise_docs/NIST_1000.$qno") or
	die "Couldn't open NIST cached documents\n";

    $topdocs = join('', <TOPDOC>);
    close(TOPDOC);

    $topdocs = demoronize($topdocs);
    $topdocs =~ s/\\[RS]//g;

    return split("Qid: $qno", $topdocs);
}
#>
#< get_documents_lucene
sub get_documents_lucene {
    my $qno = shift;

    my $topdocs = '';

    open(TOPDOC, "< /data/infolab/TREC12.FACTOID.soup/$qno.docs") or
	die "Couldn't open NIST cached documents\n";

    $topdocs = join('', <TOPDOC>);
    close(TOPDOC);

    $topdocs = demoronize($topdocs);
    $topdocs =~ s/\\[RS]//g;

    return split("Question: $qno", $topdocs);
}
#>
#< get_documents_trec2004
sub get_documents_trec2004 {
    my $qno = shift;

    my $topdocs = '';

    open(TOPDOC, "< /data/infolab/trec2004-projection-docs/$qno.txt") or
	die "Couldn't open cached documents\n";

    $topdocs = join('', <TOPDOC>);
    close(TOPDOC);

    $topdocs = demoronize($topdocs);
    $topdocs =~ s/\\[RS]//g;

    return split(/Rank:\s+\d+\s+/, $topdocs);
}
#>

#< project
sub project {
    my ($query, $ans, $docs) = @_;

    return unless $ans;

    print "Projecting '$ans'\n";

    my $optional = '';
    while ( $ans =~ /<<(.*?)>>/g ) {
	$optional .= "$1 ";
    }
    $optional =~ s/ +$//;

    $ans =~ s/<<.*?>>//g;
    $ans =~ s/^ +//g;

    #print "[$optional]\n";
    #print "[$ans]\n";

    # clever code to tokenize, drop stopwords, and strip punctuation
    my @optional_toks = grep { not is_stopword($_) }
      map { s/^\W+//; s/\W+$//; $_} split(/[ \-\/]+/, $optional);

    my @ans_toks = grep { not is_stopword($_) }
      map { s/'/'?/g; s/^\W+//; s/\W+$//; $_} split(/[ \-\/+]/, $ans);

    my @query_toks = grep { not is_stopword($_) }
      map { s/^\W+//; s/\W+$//; $_} split(/[ \-\/]+/, $query);

    # expand feet
    for my $i ( 0 .. $#{@ans_toks} ) {
	$ans_toks[$i] =~ s/^ft$/(?:ft\\.?|feet|foot)/;
	#$ans_toks[$i] =~ s/^(\d)(\d)\d\d$/(?:$1,?$2\\d\\d)/;
	#$ans_toks[$i] =~ s/^(\d\d)\d\d\d$/(?:$1,?\\d\\d\\d)/;
	#$ans_toks[$i] =~ s/^(\d),(\d)\d\d$/(?:$1,?$2\\d\\d)/;
    }

    print "Answer Tokens: @ans_toks\n";
    print "Optional Answer Tokens: @optional_toks\n";
    print "Query Tokens: @query_toks\n";

    my $max = {};
    $max->{score} = 0;
    $max->{docno} = '';

    for my $i ( 0 .. $#{@$docs} ) {
	my $text = @$docs[$i];

	next unless $text;

	# get docno from either Lucene or PRISE documents
	unless ( $text =~ /<DOCNO>\s*([^ <]+)/s or
	         $text =~ /Document Number: ([^\n ]+)/s or
	         $text =~ /Docno: ([^\n ]+)/s ) {
	    warn "Warning: No DOCNO for doc $i at query '$query'\n";
	    next;
	}

	my $docno = $1;
	$docno =~ s/\W+$//g;

	$text =~ s/<[^>]+>//g;
	$text =~ s/\n/ /g;
	$text =~ s/ +/ /g;

	my $width = 70;
	#print "Length = " . length($text) . "\n";
	my $offset = 0;
	while ( $offset < length($text) ) {
	    my $window = substr($text, $offset, $width*2);
	    #print "$offset==$window\n";
	    my $score = 0;
	    my $new_ans = $ans;

	    if ( $ans =~ /\d\d\d,\d\d\d/ ) {
		($score, $new_ans) = score_large_numbers($window, $query, $ans, \@ans_toks, \@query_toks);
	    } elsif ( $query =~ /^When/i ) {
		($score, $new_ans) = score_when($window, $ans, \@ans_toks, \@query_toks);
	    } elsif ( $query =~ /^Where/i ) {
		($score, $new_ans) = score_where($window, $ans, \@ans_toks, \@optional_toks, \@query_toks);
	    } elsif ( $query =~ /^Who/i ) {
		($score, $new_ans) = score_who($window, $ans, \@ans_toks, \@query_toks);
	    } else {
		($score, $new_ans) = score_default($window, $ans, \@ans_toks, \@query_toks);
	    }

	    if ( $score > $max->{score} and $score > 0) {
		$max->{score} = $score;
		$max->{docno} = $docno;
		$max->{ans} = $new_ans;
		print "==$score==$new_ans==$window==\n";
	    }

	    $offset += ($width * 0.5);
	}
    }

    return ($max->{docno}, $max->{score}, $max->{ans});
}
#>

#< score_large_numbers
sub score_large_numbers {
    my ($window, $query, $ans, $ans_toks, $query_toks) = @_;

    my $score = 0;
    $ans =~ /[\d,]+/;
    my $target = $&;
    $target =~ s/,//g;

    $score ++ if $query =~ /population/ and $window =~ /people/;

    $score += grep { $window =~ /(^| |\W)\Q$_\E/i or
		     $window =~ /\Q$_\E( |\W|$)/i
		   } @$query_toks;

    # essentially, double the score of proper nouns
    $score += grep { $window =~ /(^| |\W)\Q$_\E/i or
		     $window =~ /\Q$_\E( |\W|$)/i
		   }
	      grep { /^[A-Z]/ } @$query_toks;

    return (0, '') unless $score;

    $score *= 15;

    #print "==$window==\n" if $window =~ /population of 15 million/;

    my $num;
    my $original;
    if ( $window =~ /(^| )\d\d\d(,\d\d\d)+/ ) {
	$num = $&;
	$original = $num;
	$num =~ s/,//g;
    }

    if ( $window =~ /(^| )(\d+(\.\d+)?) (million|billion|trillion)/ ) {
	my ($val, $mul) = ($2, $4);
	$original = $&;
	$original =~ s/^\W+//g;

	$val *= 1000000 if $mul eq 'million';
	$val *= 1000000000 if $mul eq 'billion';
	$val *= 1000000000000 if $mul eq 'billion';
	$num = $val;

	#print "==$target,$val==\n" if $window =~ /population of 15 million/;
	#$num = $&;
	#$num =~ s/,//g;
    }

    if ( $num ) {
	my $diff = abs($target-$num)/$target;
	#print "==$score,$diff==\n" if $window =~ /population of 15 million/;
	if ( $diff < 0.5 ) {
	    #print "$num $diff $score " . ((1-$diff) * 100) . "\n";
	    #print "$window\n";
	    $score += ((1-$diff) * 100);

	    return ($score, $original);
	}
    }

    return (0, '');
}
#>
#< score_when
sub score_when {
    my ($window, $ans, $ans_toks, $query_toks) = @_;

    my $score = 0;
    # if dates match exactly, and the answer isn't a single token,
    # then most likely we've got it correctly.
    my $one_shot = $ans;
    $one_shot =~
	s/(^| )($MONTH_FULL)(\W|$)/$1($2|$MONTH_TO_ABBREV{$2}\\.?)$3/g;
    if ( $#{@$ans_toks} > 0 and $window =~ /$one_shot/ ) {
	$score = ($#{@$ans_toks} + 1) * 5;
    } else {
	foreach my $tok ( @$ans_toks ) {
	    next if is_stopword($tok);

	    # search for abbreviation of months also
	    $tok =~ s/^($MONTH_FULL)$/($1|$MONTH_TO_ABBREV{$1}\\.?)/g;

	    if ( $tok =~ /\d\d\d\d/ and $window =~ /$tok/ ) {
		$score += 5;
	    }
	    elsif ( $tok =~ /\d+,?/ and $window =~ /$tok/ ) {
		$score += 1;
	    }
	    elsif ( $window =~ /$tok/ ) {
		$score += 2;
	    }
	}
    }

    return (0, $ans) if $score == 0;

    foreach my $tok ( @$query_toks ) {
	next if is_stopword($tok);
	if ( $window =~ /$tok/ ) {
	    $score += get_idf($tok);
	}
    }

    # Sometimes, the corpus answer is actually more exact than the
    # Aranea answer.  In those cases, take the corpus answer.
    if ( $ans =~ /^${YEAR}$/ ) {
	return ($score, $&) if $window =~ /$MONTH,?\s+$ans/;
	return ($score, $&) if $window =~ /$MONTH\s+$DAY,?\s+$ans/;
    }

    if ( $ans =~ /^$MONTH $YEAR$/ ) {
	return ($score, $&) if $window =~ /$MONTH\s+$DAY,?\s+$YEAR/;
    }

    return ($score, $ans) if $window =~ /\Q$ans\E/;

    # If the corpus answer is less exact than the Aranea answer, use
    # the corpus answer.
    if ( $ans =~ /($MONTH) ($DAY), ($YEAR)/ ) {
	my ($month, $day, $year) = ($1, $2, $3);

	my $abbrev = $MONTH_TO_ABBREV{$month};
	my $a = "$abbrev\\.? $day,? $year";
	# Months are often abbreviated
	return ($score, $ans)
	    #if $window =~ /$abbrev\\.? $day,? $year/;
	    if $window =~ /$a/;

	# April 18 is the anniversary of the great San Francisco
	# earthquake of 1906, which triggered fires which
	# destroyed most...
	return ($score, $ans)
	    if $window =~ /$month $day/ and $window =~ /$year/;

	return ($score, "##$month, $year")
	    if $window =~ /$month,?\s+$year/ or
	       $window =~ /$abbrev\.?\s+$year/;

	return ($score, "$year") if $window =~ /$year/;

    } elsif ( $ans =~ /($MONTH),?\s+($YEAR)/ ) {
	my ($month, $year) = ($1, $2);

	return ($score, "$month, $year") if $window =~ /$month,?\s+$year/;
	return ($score, "$year") if $window =~ /$year/;
    }

    my @toks = map { s/'/'?/g; s/^\W+//; s/\W+$//; $_}
	split(/[ \-\/+]/, $ans);

    my $new_ans = '';
    foreach my $tok ( @toks ) {
	if ( $window =~ /(^| |\W)$tok( |\W|$)/i ) {
	    $new_ans .= "$tok ";
	}
    }
    $new_ans =~ s/ $//g;

    return ($score, $new_ans);
}
#>
#< score_where
sub score_where {
    my ($window, $ans, $ans_toks, $optional_toks, $query_toks) = @_;

    my $score = 0;

    # match abbreviations of full state names
    if ( $ans =~ /$STATE$/ ) {
	my $tok = $&;
	#print "##$tok\n";

	my $abbrev = $STATE_TO_ABBREV{$tok}->[0];
	#print "##$abbrev\n";
	$score += get_idf($tok)
	    if $window =~ /(^| |\W)\Q$abbrev\E( |\W|$)/;
    }

    foreach my $tok ( @$ans_toks ) {
	if ( $window =~ /(^| |\W)\Q$tok\E( |\W|$)/ ) {
	    $score += get_idf($tok);
	} elsif ( $window =~ /(^| |\W)\Q$tok\E( |\W|$)/i ) {
	    $score += (0.5 * get_idf($tok));
	}
    }

    foreach my $tok ( @$optional_toks ) {
	if ( $window =~ /(^| |\W)\Q$tok\E( |\W|$)/ ) {
	    $score += (0.5 * get_idf($tok));
	} elsif ( $window =~ /(^| |\W)\Q$tok\E( |\W|$)/i ) {
	    $score += (0.25 * get_idf($tok));
	}
    }

    return (0, '') if $score == 0;

    foreach my $tok ( @$query_toks ) {
	if ( $window =~ /(^| |\W)\Q$tok\E( |\W|$)/ ) {
	    $score += get_idf($tok);
	} elsif ( $window =~ /(^| |\W)\Q$tok\E( |\W|$)/i ) {
	    $score += (0.5 * get_idf($tok));
	}
    }

    my $new_ans = '';

    if ( $ans =~ /^Washington, DC$/ ) {
	# special case: do nothing (we don't want to drop the DC by
	# accident)
	return ($score, $ans);
    } elsif ( $ans =~ /((?:[A-Z][^ ]+ )*[A-Z][^,]+), ($STATE)/o ) {
	# treat "city, state" answers as a special case
	my ($city, $state) = ($1, $2);

	# This is a last ditch effort: sometimes, there are extra
	# words tacked on in front of the real location name, e.g.,
	# q464 'Corporate Center Indianapolis, Indiana'.  To fix this,
	# try projecting only the final token of the city.
	my $alt_city = '';
	$alt_city = $1 if $city =~ / (\w+)$/;
	# but don't project 'County', as in 'Hardin County, Kentucky'
	$alt_city = '' if $alt_city eq 'County';

	my $has_city = 0;
	my $has_state = 0;
	my $has_alt_city = 0;

	$has_state = 1
	    if $window =~/(^| |\W)$state( |\W|$)/i;
	$has_state = 1
	    if $window =~/(^| |\W)$STATE_TO_ABBREV{$state}->[0]( |\W|$)/i;
	$has_state = 1
	    if $window =~/(^| |\W)$STATE_TO_POSTAL_CODE{$state}->[0]( |\W|$)/i;

	$has_city = 1 if $window =~ /(^| |\W)$city( |\W|$)/i;
	$has_alt_city = 1 if $alt_city and $alt_city =~ /(^| |\W)$alt_city( |\W|$)/;

	return ($score, "$city, $state") if $has_city and $has_state;
	return ($score, "$alt_city, $state") if $has_alt_city and $has_state;
	return ($score, "$state") if $has_state;
	return ($score, "$city") if $has_city;
	return ($score, "$alt_city") if $has_alt_city;
	return ($score, '');
    } if ( $window =~ /$ans/ and $#{@$optional_toks} == -1) {
	# return if we have an exact match and there are no optional
	# tokens to match; this is needed for answers like q1326
	# 'Tower of London', where stopwords are essential.
	return ($score, $ans);
    }

    if ( $#{@$optional_toks} != -1 ) {
	foreach my $tok ( @$optional_toks ) {
	    if ( $window =~ /(^| |\W)$tok( |\W|$)/i ) {
		$new_ans .= "$tok ";
	    }
	}

	foreach my $tok ( @$ans_toks ) {
	    if ( $window =~ /(^| |\W)$tok( |\W|$)/i ) {
		$new_ans .= "$tok ";
	    }
	}

	$new_ans =~ s/ $//g;
    } else {
	my @toks = map { s/'/'?/g; s/^\W+//; s/\W+$//; $_}
	    split(/[ \-\/+]/, $ans);

	foreach my $tok ( @toks ) {
	    if ( $window =~ /(?:^| |\W)($tok)(?: |\W|$)/i ) {
		$new_ans .= "$1 ";
	    }
	}
	$new_ans =~ s/ $//g;
    }

    return ($score, $new_ans);
}

#>
#< score_who
sub score_who {
    my ($window, $ans, $ans_toks, $query_toks) = @_;

    my $score = 0;
    foreach my $tok ( @$ans_toks ) {
	if ( $window =~ /(^| |\W)\Q$tok\E( |\W|$)/ ) {
	    $score += get_idf($tok);
	} elsif ( $window =~ /(^| |\W)\Q$tok\E( |\W|$)/i ) {
	    $score += (0.5 * get_idf($tok));
	}
    }

    return (0, '') if $score == 0;

    foreach my $tok ( @$query_toks ) {
	if ( $window =~ /(^| |\W)\Q$tok\E( |\W|$)/ ) {
	    $score += get_idf($tok);
	} elsif ( $window =~ /(^| |\W)\Q$tok\E( |\W|$)/i ) {
	    $score += (0.5 * get_idf($tok));
	}
    }

    return ($score, $ans) if $window =~ /$ans/;

    my @toks = map { s/'/'?/g; s/^\W+//; s/\W+$//; $_}
	split(/[ \-\/+]/, $ans);

    my $new_ans = '';
    foreach my $tok ( @toks ) {
	if ( $window =~ /(^| |\W)$tok( |\W|$)/i ) {
	    $new_ans .= "$tok ";
	}
    }
    $new_ans =~ s/ $//g;

    #$new_ans =~ s/ $//g;
    #
    #my $new_ans = '';
    #foreach my $tok ( @$ans_toks ) {
	#if ( $window =~ /(^| |\W)$tok( |\W|$)/i ) {
	#    $new_ans .= "$tok ";
	#}
    #}

    #$new_ans =~ s/ $//g;

    # if we were only able to find support for the first name, return
    # the entire answer string... maybe we'll get lucky
    return ($score, $ans)
	if $new_ans =~ /($FEMALE_FIRST_NAME|$MALE_FIRST_NAME)/ and
	    $ans =~ /^$new_ans/;

    return ($score, $new_ans);
}

#>
#< score_default
sub score_default {
    my ($window, $ans, $ans_toks, $query_toks) = @_;

    my $score = 0;
    foreach my $tok ( @$ans_toks ) {
	if ( $window =~ /(^| |\W)\Q$tok\E( |\W|$)/ ) {
	    $score += get_idf($tok);
	} elsif ( $window =~ /(^| |\W)\Q$tok\E( |\W|$)/i ) {
	    $score += (0.5 * get_idf($tok));
	}
    }

    return (0, '') if $score == 0;

    foreach my $tok ( @$query_toks ) {
	if ( $window =~ /(^| |\W)\Q$tok\E( |\W|$)/ ) {
	    $score += get_idf($tok);
	} elsif ( $window =~ /(^| |\W)\Q$tok\E( |\W|$)/i ) {
	    $score += (0.5 * get_idf($tok));
	}
    }

    return ($score, $ans) if $window =~ /$ans/;

    my @toks = map { s/'/'?/g; s/^\W+//; s/\W+$//; $_}
	split(/[ \-\/+]/, $ans);

    # expand feet
    for my $i ( 0 .. $#{@toks} ) {
	$toks[$i] =~ s/^ft$/(ft\\.?|feet|foot)/;
	#$toks[$i] =~ s/^(\d)(\d)\d\d$/(?:$1,?$2\\d\\d)/;
	#$toks[$i] =~ s/^(\d\d)\d\d\d$/(?:$1,?\\d\\d\\d)/;
	#$toks[$i] =~ s/^(\d),(\d)\d\d$/(?:$1,?$2\\d\\d)/;
    }

    my $new_ans = '';
    foreach my $tok ( @toks ) {
	if ( $window =~ /(?:^| |\W)($tok)(?: |\W|$)/i ) {
	    $new_ans .= "$1 ";
	}
    }
    $new_ans =~ s/ $//g;

    # if we lost the number (at the beginning), then give up and
    # return original answer
    return ($score, $ans) if $ans =~ /^\d/ and $new_ans !~ /\d/;

    return ($score, $new_ans);
}
#>

1;

