#!/usr/bin/perl -w
#< Boilerplate
#
# The Aranea question answering system
# Copyright (C) 2002-2005, Jimmy Lin
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, please visit
# http://www.gnu.org/copyleft/gpl.html
#
# $Header: /fs/clip-qa/.cvsroot//Aranea/pm/Aranea/Modules/ScoreCandidates.pm,v 1.8 2005/06/10 16:17:16 jimmylin Exp $
#
#>
#< POD
###############################################################################

=head1 NAME

B<C<Aranea::Modules::ScoreCandidate>> - Aranea module for scoring
answer candidates based on their I<idf> values.

=head1 SYNOPSIS

This Aranea modules is meant to be used in conjunction with the
L<C<Aranea::Shell>|Aranea::Shell>.

=head1 DESCRIPTION

This Aranea module adjusts the score of each candidate answer based on
the average I<idf> values of all component unigrams.  The new score is
the product of the old score and the average I<idf> value.  Taken
together, the modules C<Vote> and C<ScoreCandidate> perform basically
a I<tfidf> weighting scheme.

=cut

###############################################################################
#>
#< Header
package Aranea::Modules::ScoreCandidates;

use strict;
use warnings;
use lib qw(lib pm);                         # needed for command-line mode

use IO::Handle;

use Aranea::XML::Aranea;
use Aranea::Lexicon::Lexicon qw(:default);

#>

#< STDIN handler
unless ( caller() ) {
    my $stdin = new_from_fd IO::Handle(fileno(STDIN),"r");
    my $Aranea = Aranea::XML::Aranea->new($stdin);

    $Aranea = doModule($Aranea);

    $Aranea->write_xml();
}
#>

#< sub doModule
sub doModule {
    my $Aranea = shift;

    $Aranea->insert_pipeline_tag('$Id: ScoreCandidates.pm,v 1.8 2005/06/10 16:17:16 jimmylin Exp $ ');

    # if there are no requests, then there is nothing to do
    return $Aranea if $Aranea->request_list()->entries_count() == -1;

    my $entries = $Aranea->candidate_list()->entries();
    for my $entry( @$entries ) {
	my $candidate = $entry->candidate();
	$entry->score($entry->score() * _score_ngram($candidate));
    }

    return $Aranea;
}

#>

#< sub _score_ngram
sub _score_ngram {
    my $ngram = shift;
    my $score = 0;
    my $cnt = 0;

    my @words = split(/\s+/, $ngram);

    foreach my $myword (@words) {
	$myword =~ s/\d/\#/g if $myword =~ /\d\d\d/g;

	$myword =~ s/\W+$//g;
	$myword =~ s/^\W+//g;

	$score += get_idf($myword);
	$cnt++;
    }

    # Experiments show that averaging the idf values gives better
    # performance than summing the idf values.  The latter technique
    # biases longer candidates, which is not a good idea; many
    # perfectly fine answers are very short.

    $score /= $cnt if $cnt != 0;
    return $score;
}
#>

1;
