#!/usr/bin/perl -w
#< Boilerplate
#
# The Aranea question answering system
# Copyright (C) 2002-2005, Jimmy Lin
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, please visit
# http://www.gnu.org/copyleft/gpl.html
#
# $Header: /fs/clip-qa/.cvsroot//Aranea/pm/Aranea/XML/CandidateEntry.pm,v 1.10 2005/06/08 01:41:46 jimmylin Exp $
#
#>
#< POD
###############################################################################

=head1 NAME

B<C<Aranea::XML::CandidateEntry>> - object representing an entry in the B<C<E<lt>candidate_listE<gt>>> of an Aranea XML object

=head1 SYNOPSIS

  my $candidate_list = $Aranea->candidate_list();
  my $entries = $candidate_list->entries();
 
  for my $entry ( @$entries ) {
     # do stuff...
  }

=head1 DESCRIPTION

The C<CandidateEntry> object represents a single entry inside the
C<candidate_list> section of an Aranea XML object.  Here is an example:

  <entry>
    <support>
      <doc> http://www.foo.com/url1.html </doc>
      <doc> http://www.bar.com/url2.html </doc>
    </support>
    <score>63</score>
    <candidate>Central America</candidate>
  </entry>

The L<C<CandidateList>|Aranea::XML::CandidateList> object serves as
a container for a collection of C<CandidateEntry>s.

=head1 METHODS

=over 4

=cut

###############################################################################
#>
#< Header
package Aranea::XML::CandidateEntry;

use strict;
use warnings;

use XML::Simple;
use XML::Writer;

#>

#< sub new

=item B<C<new($candidate, $score)>>

Constructs a C<CandidateEntry> given a candidate and a score
for that candidate.

=cut

# August 3, 2002 -- Note on AUTOLOADing: instead dynamically
# generating the basic accessor and mutators on the fly, I have
# choosen to manually define them explictly.  Previously, candidate()
# and score(), two often used methods, were AUTOLOADed; profiling with
# Devel::DProf revealed that 22% of all cpu cycles were spent inside
# AUTOLOAD.  Explictly defining the methods dramatically speeds up the
# application.
sub new {
    my $proto = shift;
    my $class = ref($proto) || $proto;
    my $self = {};

    # Secretly, this is a overloaded constructor (i.e., not mentioned
    # in the POD).  If only one arg is pased to the constructor, then
    # we are creating a CandidateEntry from XML input.  Otherwise, we
    # are given the candidate itself and the score.

    if ( $#{@_} == 1 ) {
	# candidate and score can be passed in through the object constructor
	$self->{candidate} = shift;
	$self->{score} = shift;
    } elsif ( $#{@_} == 0 ) {
	$self = shift;
    }

    bless ($self, $class);
    return $self;
}
#>

#< sub candidate

=item B<C<candidate([$candidate])>>

Accesses or mutates the candidate answer.

=cut
sub candidate {
    my $self = shift;

    if (@_) {
	return $self->{candidate} = shift;
    }
    else {
	return $self->{candidate};
    }
}
#>
#< sub score

=item B<C<score([$score])>>

Accesses or mutates the score of the candidate answer.

=cut
sub score {
    my $self = shift;

    if (@_) {
	return $self->{score} = shift;
    }
    else {
	return $self->{score};
    }
}
#>
#< sub projection_score

=item B<C<projection_score([$score])>>

Accesses or mutates the projection score of the candidate answer.

=cut
sub projection_score {
    my $self = shift;

    if (@_) {
	return $self->{projection_score} = shift;
    }
    else {
	return $self->{projection_score};
    }
}
#>
#< sub support

=item B<C<support([$support])>>

Accesses or mutates the support of the candidate answer.  This method
of access is not preferred because it returns a hashref, which
directly reflects the underlying XML structure.  Instead, use the
C<support_docs> method.

=cut
sub support {
    my $self = shift;

    if (@_) {
	return $self->{support} = shift;
    }
    else {
	return $self->{support};
    }
}
#>
#< sub add_support

=item B<C<add_support($url)>>

Adds a URL to the list of supporting documents.

=cut
sub add_support {
    my $self = shift;
    my $support = shift;

    push @{$self->{support}->{doc}}, $support;
}
#>
#< sub support_count

=item B<C<support_count($url)>>

Returns the number of supporting URLs.  Note that the return value is
similiar to that of C<$#>, i.e., -1 for no entries, 0 for one entry,
etc.

=cut
sub support_count {
    my $self = shift;

    return $#{@{$self->{support}->{doc}}};
}
#>
#< sub support_docs

=item B<C<support_docs([$listref])>>

Accesses or mutates the supporting URLs.  When called with no
arguments, this method returns a listref of supporting URLs.  When
called with a single listref, the supporting URLs is set to the
listref.

=cut
sub support_docs {
    my $self = shift;

    @_ ? $self->{support}->{doc} = shift : $self->{support}->{doc};
}
#>
#< sub write_xml

=item B<C<write_xml($writer)>>

Writes the XML representation of this C<CandidateList>; takes an
C<XML::Writer> as its only argument.  Note that this method is rarely
invoked manually; under typical usage patterns, the C<write_xml>
method of L<C<Aranea::XML::Aranea>|Aranea::XML::Aranea> will call this
method automatically.

=cut
sub write_xml {
    my ($self, $writer) = @_;

    $self->_recursive_write($self, $writer);
}
#>

#< sub _recursive_write
# NOTE: this is a "private" method; therefore, no POD.
#
# This method knows how to handle the printing of a generic XML
# object.  If an element is embedded inside another one, a recursive
# call is made to itself.  If an element is actually a list, a
# recursive call is mapped over the list.  For the base case, the
# element is merely printed.
sub _recursive_write {
    my ($self, $entry, $writer) = @_;

    my @keys = keys(%$entry);             # get all the elements

    for my $akey (@keys) {                # iterate through all elements
	next if $akey =~ /^_/;            # ignore private elements

	my $struc = $entry->{$akey};

	if ( ref($struc) eq "HASH" ) {
	    # embedded element, recurse.
	    $writer->startTag($akey);
	    $self->_recursive_write($struc, $writer);
	    $writer->endTag();
	}
	elsif ( ref($struc) eq "ARRAY" ) {
	    # list, map recurse over all elements in array
	    foreach my $elt ( @$struc ) {
		if ( ref($elt) ) {
		    # if each element in the list is a reference
		    # itself, then recurse
		    $writer->startTag($akey);
		    recursive_write($struc, $writer);
		    $writer->endTag();
		} else {
		    # otherwise just print out the element
		    $writer->startTag($akey);
		    $writer->characters($elt);
		    $writer->endTag();
		}
	    }
	}
	else {
	    # base case: merely print out the element
	    $writer->startTag($akey);
	    $writer->characters($struc);
	    $writer->endTag();

	}
    }
}
#>

1;
