#!/usr/bin/perl -w
#< Boilerplate
#
# The Aranea question answering system
# Copyright (C) 2002-2005, Jimmy Lin
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, please visit
# http://www.gnu.org/copyleft/gpl.html
#
# $Header: /fs/clip-qa/.cvsroot//Aranea/pm/Aranea/XML/Query.pm,v 1.9 2005/06/08 01:41:47 jimmylin Exp $
#
#>
#< POD
###############################################################################

=head1 NAME

B<C<Aranea::XML::Query>> - object representing the B<C<E<lt>queryE<gt>>> in an Aranea XML object

=head1 SYNOPSIS

  use Aranea::XML::Aranea;
  use Aranea::XML::Query;
 
  my $stdin = new_from_fd IO::Handle(STDIN, "r");
  my $Aranea = Aranea::XML::Aranea->new($stdin);
 
  my $query = $Aranea->query();
  my $text = $query->text();      # text of the question
  my $id = $query->id();          # the question id (number)

=head1 DESCRIPTION

The C<PageList> object represents the C<page_list> section of an
Aranea XML object.  Here is an example:

  <query>
    <text>What is the volcano that destroyed the ancient city of Pompeii?</text>
    <id>1396</id>
    <original>What is the name of the volcano that destroyed the ancient city of Pompeii?</original>
    <tagged>What/WP is/VBZ the/DT volcano/NN that/WDT destroyed/VBD the/DT ancient/JJ city/NN of/IN Pompeii/NNP ?/. </tagged>
  </query>

Under normal circumstances, the creation of a C<Query> object should
be handled automatically by the
L<C<Aranea::XML::Aranea>|Aranea::XML::Aranea> object.

=head1 METHODS

=over 4

=cut

###############################################################################
#>
#< Header
package Aranea::XML::Query;

use strict;
use warnings;

use XML::Simple;
use XML::Writer;

our $AUTOLOAD;

#>

#< defined fields
my %fields = (
	      text => undef,
	      id => undef,
	      tagged => undef,
	      original => undef,
	      qfocus => undef,
	      );
#>
#< sub new

=item B<C<new($data)>>

Constructs a C<Query> object given parsed XML data.  Note that this
method is rarely invoked manually; under typical usage patterns, the
constructor of L<C<Aranea::XML::Aranea>|Aranea::XML::Aranea> will call
this method automatically.

=cut
sub new {
    my $proto = shift;
    my $data = shift;
    my $class = ref($proto) || $proto;
    my $self = $data;

    $self->{_permitted} = \%fields,

    bless ($self, $class);
    return $self;
}
#>
#< sub AUTOLOAD

=item B<C<text([$str])>>

Accesses or mutates the text of the query.

=item B<C<id([$str])>>

Accesses or mutates the question number or id.

=item B<C<tagged([$str])>>

Accesses or mutates the part-of-speech tagged question.

=item B<C<original([$str])>>

Accesses or mutates the original (unmodified) text of the question.

=item B<C<qfocus([$str])>>

Accesses or mutates the question focus.

=cut
sub AUTOLOAD {
    my $self = shift;
    my $type = ref($self)
	or die "$self is not an object";

    my $name = $AUTOLOAD;
    $name =~ s/.*://;   # strip fully-qualified portion
    if ($name =~ m/DESTROY/) {
	return;
    }

    unless (exists $self->{_permitted}->{$name} ) {
	die "Can't access `$name' field in class $type";
    }

    if (@_) {
	return $self->{$name} = shift;
    }
    else {
	return $self->{$name};
    }
}
#>

#< sub add_reformulation
sub add_reformulation {
  my $self = shift;
  my $reform = shift;

  push @{$self->{reformulations}->{reformulation}}, $reform;
}
#>
#< sub get_reformulation
sub get_reformulations {
  my $self = shift;

  return $self->{reformulations}->{reformulation};
}
#>
#< sub write_xml

=item B<C<write_xml($writer)>>

Writes the XML representation of this C<PageList>; takes an
C<XML::Writer> as its only argument.  Note that this method is rarely
invoked manually; under typical usage patterns, the C<write_xml>
method of L<C<Aranea::XML::Aranea>|Aranea::XML::Aranea> will call this
method automatically.

=cut
sub write_xml {
    my $self = shift;
    my $writer = shift;

    $writer->startTag('query');
    for my $k ( keys %fields ) {
	next unless defined $self->{$k};

	$writer->startTag($k);
	$writer->characters($self->{$k});
	$writer->endTag();
    }
    $writer->endTag();
}
#>

1;
