#!/usr/bin/perl -w
#< Boilerplate
#
# The Aranea question answering system
# Copyright (C) 2002-2005, Jimmy Lin
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, please visit
# http://www.gnu.org/copyleft/gpl.html
#
# $Header: /fs/clip-qa/.cvsroot//Aranea/pm/Aranea/XML/RequestEntry.pm,v 1.6 2005/06/08 01:41:47 jimmylin Exp $
#
#>
#< POD
###############################################################################

=head1 NAME

B<C<Aranea::XML::RequestEntry>> - object representing an entry in the B<C<E<lt>request_listE<gt>>> of an Aranea XML object

=head1 SYNOPSIS

  my $request_list = $Aranea->request_list();
  my $entries = $request_list->entries();
 
  for my $request ( @$entries ) {
     # do stuff...
  }

=head1 DESCRIPTION

The C<RequestEntry> object represents a single entry inside the
C<candidate_list> section of an Aranea XML object.  Here is an example:

  <entry>
    <query>What is the volcano that destroyed the ancient city of Pompeii?</query>
    <score>1</score>
    <constraints>
      <use_backoff>true</use_backoff>
      <pages_to_search>100</pages_to_search>
    </constraints>
  </entry>

The L<C<RequestList>|Aranea::XML::RequestList> object serves as
a container for a collection of C<RequestEntry>s.

=head1 METHODS

=over 4

=cut

###############################################################################
#>
#< Header
package Aranea::XML::RequestEntry;

use strict;
use warnings;

use XML::Simple;
use XML::Writer;

#>

#< sub new

=item B<C<new($data)>>

Constructs a C<Request> object given parsed XML data.  Note that the
preferred way of creating requests is to use the
C<add_default_exact_query> or C<add_default_inexact_query> methods of
L<C<RequestList>|Aranea::XML::RequestList>, which creates a query with
the default parameters:

  $request_list->add_default_exact_query("?x killed Lincoln");

=cut
sub new {
    my $proto = shift;
    my $data = shift;
    my $class = ref($proto) || $proto;
    my $self = $data;

    bless ($self, $class);
    return $self;
}
#>

#< sub query

=item B<C<query([$request])>>

Accesses or mutates the query.

=cut
sub query {
    my $self = shift;

    if (@_) {
	return $self->{query} = shift;
    }
    else {
	return $self->{query};
    }
}
#>
#< sub score

=item B<C<score([$request])>>

Accesses or mutates the score.

=cut
sub score {
    my $self = shift;

    if (@_) {
	return $self->{score} = shift;
    }
    else {
	return $self->{score};
    }
}
#>
#< sub get_constraint

=item B<C<get_constraint($constraint)>>

Returns a specific constraint.

=cut
sub get_constraint {
    my $self = shift;
    my $c = shift;

    return $self->{constraints}->{$c};
}
#>
#< sub write_xml

=item B<C<write_xml($writer)>>

Writes the XML representation of this C<CandidateList>; takes an
C<XML::Writer> as its only argument.  Note that this method is rarely
invoked manually; under typical usage patterns, the C<write_xml>
method of L<C<Aranea::XML::Aranea>|Aranea::XML::Aranea> will call this
method automatically.

=cut
sub write_xml {
    my $self = shift;
    my $writer = shift;

    $writer->startTag('entry');
    $writer->startTag('query');
    $writer->characters($self->{query});
    $writer->endTag();

    $writer->startTag('score');
    $writer->characters($self->{score});
    $writer->endTag();

    $writer->startTag('constraints');

    my @ckeys = keys(%{$self->{constraints}});

    for my $k (@ckeys) {
	$writer->startTag($k);
	$writer->characters($self->{constraints}->{$k});
	$writer->endTag();
    }

    $writer->endTag();
    $writer->endTag();
}
#>

#< More POD
###############################################################################

=back

=head1 COPYRIGHT

The Aranea Question Answering Project.
Copyright (C) 2002 by Jimmy Lin (jimmylin@ai.mit.edu)

=cut

###############################################################################
#>

1;
